"""Collects required template histograms provided by user."""

import logging
import pathlib
from typing import Any, cast, Dict, Optional

from cabinetry import histo
from cabinetry import route
from cabinetry._typing import Literal
from cabinetry.templates import utils


log = logging.getLogger(__name__)


def _histo_path(
    general_path: str,
    variation_path: str,
    region: Dict[str, Any],
    sample: Dict[str, Any],
    systematic: Dict[str, Any],
    template: Optional[Literal["Up", "Down"]],
) -> str:
    """Returns the paths to a histogram for a region-sample-systematic-template.

    The path should contain a colon to separate the path to the file itself and the
    location of the histogram within the file. Due to the presence of this colon, the
    return value is a string instead of a pathlib path. A path is built starting from
    the path specified in the general options in the configuration file. This path
    contains placeholders for region-, sample-, and systematic-specific values.

    Args:
        general_path (str): path specified in general settings, with sections that can
            be overridden by region / sample / systematic settings
        variation_path (str): default for VariationPath setting
        region (Dict[str, Any]): containing all region information
        sample (Dict[str, Any]): containing all sample information
        systematic (Dict[str, Any]): containing all systematic information
        template (Optional[Literal["Up", "Down"]]): template considered: "Up", "Down",
            or None for nominal

    Raises:
        ValueError: when ``RegionPath`` placeholder is used, but region setting is not
            specified
        ValueError: when ``SamplePath`` placeholder is used, but sample setting is not
            specified

    Returns:
        str: path to input file (and histogram within)
    """
    # obtain region and sample paths, if they are defined
    region_path = region.get("RegionPath", None)
    sample_path = sample.get("SamplePath", None)

    # check whether a systematic is being processed
    if template is not None:
        # determine whether the template has an override for VariationPath specified
        variation_override = utils._check_for_override(
            systematic, template, "VariationPath"
        )
        if variation_override is not None:
            # _check_for_override should return Optional[str] for VariationPath, but
            # mypy cannot know that it will not be a list, so explicitly cast to str
            variation_path = cast(str, variation_override)
        else:
            log.warning(
                f"no VariationPath override specified for {region['Name']} / "
                f"{sample['Name']} / {systematic['Name']} {template}"
            )
    # apply variation-specific setting
    path = general_path.replace("{VariationPath}", variation_path)

    # handle region-specific setting
    region_template_exists = "{RegionPath}" in general_path
    if region_path is not None:
        if not region_template_exists:
            log.warning(
                "region override specified, but {RegionPath} not found in default path"
            )
        path = path.replace("{RegionPath}", region_path)
    elif region_template_exists:
        raise ValueError(f"no path setting found for region {region['Name']}")

    # handle sample-specific setting
    sample_template_exists = "{SamplePath}" in general_path
    if sample_path is not None:
        if not sample_template_exists:
            log.warning(
                "sample override specified, but {SamplePath} not found in default path"
            )
        path = path.replace("{SamplePath}", sample_path)
    elif sample_template_exists:
        raise ValueError(f"no path setting found for sample {sample['Name']}")

    # check for presence of colon to distinguish path to file and location within file
    if ":" not in path:
        log.warning(f"no colon found in path {path}, may not be able to find histogram")

    return path


def _collector(
    histogram_folder: pathlib.Path, general_path: str, variation_path: str, method: str
) -> route.ProcessorFunc:
    """Returns the histogram-collecting function to be applied to template histograms.

    Needed by ``cabinetry.route.apply_to_all_templates``. Could alternatively create a
    ``Collector`` class that contains processors (see ``builder._Builder`` for an
    example).

    Args:
        histogram_folder (pathlib.Path): folder to save the histograms to
        general_path (str): template for paths to input files for histogram building
        variation_path (str): default for VariationPath setting
        method (str): backend to use for histogram production

    Raises:
            NotImplementedError: when requesting an unknown backend

    Returns:
        route.ProcessorFunc: function to apply to a template histogram
    """

    def collect_template(
        region: Dict[str, Any],
        sample: Dict[str, Any],
        systematic: Dict[str, Any],
        template: Optional[Literal["Up", "Down"]],
    ) -> None:
        """Collects a histogram and writes it to a file.

        Args:
            region (Dict[str, Any]): containing all region information
            sample (Dict[str, Any]): containing all sample information
            systematic (Dict[str, Any]): containing all systematic information
            template (Optional[Literal["Up", "Down"]]): template considered: "Up",
                "Down", or None for nominal
        """
        histo_path = _histo_path(
            general_path, variation_path, region, sample, systematic, template
        )

        # obtain the histogram
        if method == "uproot":
            from cabinetry.contrib import histogram_reader

            histogram = histogram_reader.with_uproot(histo_path)

        else:
            raise NotImplementedError(f"unknown backend {method}")

        # store information in a Histogram instance and save it
        utils._name_and_save(
            histogram_folder,
            histo.Histogram(histogram),
            region,
            sample,
            systematic,
            template,
        )

    return collect_template



import logging
from unittest import mock

import pyhf
import pytest

from cabinetry import model_utils
from cabinetry import tabulate


@pytest.mark.parametrize(
    "test_input, expected", [(("abc", 0), "abc\nbin 1"), (("abc", 2), "abc\nbin 3")]
)
def test__header_name(test_input, expected):
    assert tabulate._header_name(*test_input) == expected
    assert tabulate._header_name("abc", 2, unique=False) == "\nbin 3"


def test__yields_per_bin(example_spec_multibin, example_spec_with_background, caplog):
    caplog.set_level(logging.DEBUG)

    # multiple channels
    model = pyhf.Workspace(example_spec_multibin).model()
    yields = [[[25.0, 5.0]], [[8.0]]]
    total_stdev = [[5.0, 2.0], [1.0]]
    data = [[35, 8], [10]]
    channels = model.config.channels
    label = "pre-fit"
    caplog.clear()

    yield_table = tabulate._yields_per_bin(
        model, yields, total_stdev, data, channels, label
    )
    assert yield_table == [
        {
            "sample": "Signal",
            "region_1\nbin 1": "25.00",
            "region_1\nbin 2": "5.00",
            "region_2\nbin 1": "8.00",
        },
        {
            "sample": "total",
            "region_1\nbin 1": "25.00 \u00B1 5.00",
            "region_1\nbin 2": "5.00 \u00B1 2.00",
            "region_2\nbin 1": "8.00 \u00B1 1.00",
        },
        {
            "sample": "data",
            "region_1\nbin 1": "35.00",
            "region_1\nbin 2": "8.00",
            "region_2\nbin 1": "10.00",
        },
    ]
    assert "yields per bin for pre-fit model prediction:" in caplog.records[0].message
    caplog.clear()

    # multiple samples
    model = pyhf.Workspace(example_spec_with_background).model()
    yields = [[[150.0], [50.0]]]
    total_stdev = [[8.60]]
    data = [[160]]
    channels = model.config.channels

    yield_table = tabulate._yields_per_bin(
        model, yields, total_stdev, data, channels, label
    )
    assert yield_table == [
        {"sample": "Background", "Signal Region\nbin 1": "150.00"},
        {"sample": "Signal", "Signal Region\nbin 1": "50.00"},
        {"sample": "total", "Signal Region\nbin 1": "200.00 \u00B1 8.60"},
        {"sample": "data", "Signal Region\nbin 1": "160.00"},
    ]


def test__yields_per_channel(
    example_spec_multibin, example_spec_with_background, caplog
):
    caplog.set_level(logging.DEBUG)

    # multiple channels
    model = pyhf.Workspace(example_spec_multibin).model()
    yields = [[30], [8.0]]
    total_stdev = [5.39, 1.0]
    data = [43, 10]
    channels = model.config.channels
    label = "pre-fit"
    caplog.clear()

    yield_table = tabulate._yields_per_channel(
        model, yields, total_stdev, data, channels, label
    )
    assert yield_table == [
        {"sample": "Signal", "region_1": "30.00", "region_2": "8.00"},
        {
            "sample": "total",
            "region_1": "30.00 \u00B1 5.39",
            "region_2": "8.00 \u00B1 1.00",
        },
        {"sample": "data", "region_1": "43.00", "region_2": "10.00"},
    ]
    assert (
        "yields per channel for pre-fit model prediction:" in caplog.records[0].message
    )
    caplog.clear()

    # multiple samples
    model = pyhf.Workspace(example_spec_with_background).model()
    yields = [[150.0, 50]]
    total_stdev = [8.60]
    data = [160]
    channels = model.config.channels

    yield_table = tabulate._yields_per_channel(
        model, yields, total_stdev, data, channels, label
    )
    assert yield_table == [
        {"sample": "Background", "Signal Region": "150.00"},
        {"sample": "Signal", "Signal Region": "50.00"},
        {"sample": "total", "Signal Region": "200.00 \u00B1 8.60"},
        {"sample": "data", "Signal Region": "160.00"},
    ]


@mock.patch("cabinetry.tabulate._yields_per_channel")
@mock.patch("cabinetry.tabulate._yields_per_bin")
@mock.patch("cabinetry.model_utils._filter_channels", side_effect=[["SR"], [], ["SR"]])
@mock.patch("cabinetry.model_utils._data_per_channel", return_value=[[12.0]])
def test_yields(mock_data, mock_filter, mock_bin, mock_channel, example_spec, caplog):
    caplog.set_level(logging.DEBUG)
    model = pyhf.Workspace(example_spec).model()
    model_pred = model_utils.ModelPrediction(model, [[[10.0]]], [[0.3]], [0.3], "pred")
    data = [12.0, 1.0]  # with auxdata to strip via mock

    tabulate.yields(model_pred, data)
    assert mock_data.call_args_list == [((model, data), {})]
    assert mock_filter.call_args_list == [((model, None), {})]
    assert mock_bin.call_args_list == [
        ((model, [[[10.0]]], [[0.3]], [[12.0]], ["SR"], "pred"), {})
    ]
    assert mock_channel.call_count == 0

    # no table to produce (does not call _filter_channels)
    tabulate.yields(model_pred, data, per_bin=False, per_channel=False)
    assert "requested neither yields per bin nor per channel, no table produced" in [
        rec.message for rec in caplog.records
    ]
    caplog.clear()

    # no channels to include
    tabulate.yields(model_pred, data, channels="abc")
    assert mock_filter.call_args == ((model, "abc"), {})
    assert mock_bin.call_count == 1  # one call from before, no new call
    assert mock_channel.call_count == 0

    # yields per channel, not per bin
    tabulate.yields(model_pred, data, per_bin=False, per_channel=True)
    assert mock_bin.call_count == 1  # one call from before
    assert mock_channel.call_args_list == [
        ((model, [[10.0]], [0.3], [12.0], ["SR"], "pred"), {})
    ]



"""Provides utilities for pyhf models."""

import json
import logging
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Union

import awkward as ak
import numpy as np
import pyhf

from cabinetry.fit.results_containers import FitResults


log = logging.getLogger(__name__)


# cache holding results from yield uncertainty calculations
_YIELD_STDEV_CACHE: Dict[Any, Tuple[List[List[float]], List[float]]] = {}


class ModelPrediction(NamedTuple):
    """Model prediction with yields and total uncertainties per bin and channel.

    Args:
        model (pyhf.pdf.Model): model to which prediction corresponds to
        model_yields (List[List[List[float]]]): yields per sample, channel and bin,
            indices: channel, sample, bin
        total_stdev_model_bins (List[List[float]]): total yield uncertainty per channel
            and per bin, indices: channel, bin
        total_stdev_model_channels (List[float]): total yield uncertainty per channel,
            index: channel
        label (str): label for the prediction, e.g. "pre-fit" or "post-fit"
    """

    model: pyhf.pdf.Model
    model_yields: List[List[List[float]]]
    total_stdev_model_bins: List[List[float]]
    total_stdev_model_channels: List[float]
    label: str


def model_and_data(
    spec: Dict[str, Any], *, asimov: bool = False, include_auxdata: bool = True
) -> Tuple[pyhf.pdf.Model, List[float]]:
    """Returns model and data for a ``pyhf`` workspace specification.

    Args:
        spec (Dict[str, Any]): a ``pyhf`` workspace specification
        asimov (bool, optional): whether to return the Asimov dataset, defaults to False
        include_auxdata (bool, optional): whether to also return auxdata, defaults to
            True

    Returns:
        Tuple[pyhf.pdf.Model, List[float]]:
            - a HistFactory-style model in ``pyhf`` format
            - the data (plus auxdata if requested) for the model
    """
    workspace = pyhf.Workspace(spec)
    model = workspace.model(
        modifier_settings={
            "normsys": {"interpcode": "code4"},
            "histosys": {"interpcode": "code4p"},
        }
    )  # use HistFactory InterpCode=4 (default in pyhf since v0.6.0)
    if not asimov:
        data = workspace.data(model, include_auxdata=include_auxdata)
    else:
        data = asimov_data(model, include_auxdata=include_auxdata)
    return model, data


def asimov_data(model: pyhf.Model, *, include_auxdata: bool = True) -> List[float]:
    """Returns the Asimov dataset (optionally with auxdata) for a model.

    Initial parameter settings for normalization factors in the workspace are treated as
    the default settings for that parameter. Fitting the Asimov dataset will recover
    these initial settings as the maximum likelihood estimate for normalization factors.
    Initial settings for other modifiers are ignored.

    Args:
        model (pyhf.Model): the model from which to construct the dataset
        include_auxdata (bool, optional): whether to also return auxdata, defaults to
            True

    Returns:
        List[float]: the Asimov dataset
    """
    asimov_data = pyhf.tensorlib.tolist(
        model.expected_data(asimov_parameters(model), include_auxdata=include_auxdata)
    )
    return asimov_data


def asimov_parameters(model: pyhf.pdf.Model) -> np.ndarray:
    """Returns a list of Asimov parameter values for a model.

    For normfactors and shapefactors, initial parameter settings (specified in the
    workspace) are treated as nominal settings. This ignores custom auxiliary data set
    in the measurement configuration in the workspace.

    Args:
        model (pyhf.pdf.Model): model for which to extract the parameters

    Returns:
        np.ndarray: the Asimov parameters, in the same order as
        ``model.config.suggested_init()``
    """
    # create a list of Asimov parameters (constrained parameters at best-fit value from
    # the aux measurement, unconstrained parameters at init specified in the workspace)
    asimov_parameters = []
    for parameter in model.config.par_order:
        if not model.config.param_set(parameter).constrained:
            # unconstrained parameter: use suggested inits (for normfactor/shapefactor)
            inits = model.config.param_set(parameter).suggested_init
        elif dict(model.config.modifiers)[parameter] in ["histosys", "normsys"]:
            # histosys/normsys: Gaussian constraint, nominal value 0
            inits = [0.0] * model.config.param_set(parameter).n_parameters
        else:
            # remaining modifiers are staterror/lumi with Gaussian constraint, and
            # shapesys with Poisson constraint, all have nominal value of 1
            inits = [1.0] * model.config.param_set(parameter).n_parameters

        asimov_parameters += inits

    return np.asarray(asimov_parameters)


def prefit_uncertainties(model: pyhf.pdf.Model) -> np.ndarray:
    """Returns a list of pre-fit parameter uncertainties for a model.

    For unconstrained parameters the uncertainty is set to 0. It is also set to 0 for
    fixed parameters (similarly to how the post-fit uncertainties are defined to be 0).

    Args:
        model (pyhf.pdf.Model): model for which to extract the parameters

    Returns:
        np.ndarray: pre-fit uncertainties for the parameters, in the same order as
        ``model.config.suggested_init()``
    """
    pre_fit_unc = []  # pre-fit uncertainties for parameters
    for parameter in model.config.par_order:
        # obtain pre-fit uncertainty for constrained, non-fixed parameters
        if (
            model.config.param_set(parameter).constrained
            and not model.config.param_set(parameter).suggested_fixed
        ):
            pre_fit_unc += model.config.param_set(parameter).width()
        else:
            if model.config.param_set(parameter).n_parameters == 1:
                # unconstrained normfactor or fixed parameter, uncertainty is 0
                pre_fit_unc.append(0.0)
            else:
                # shapefactor
                pre_fit_unc += [0.0] * model.config.param_set(parameter).n_parameters
    return np.asarray(pre_fit_unc)


def _hashable_model_key(
    model: pyhf.pdf.Model,
) -> Tuple[str, Tuple[Tuple[str, str], ...]]:
    """Compute a hashable representation of the values that uniquely identify a model.

    The ``pyhf.pdf.Model`` type is already hashable, but it uses the ``__hash__``
    inherited from ``object``, so a copy of a model has a distinct hash. The key
    returned by this function instead will hash to the same value for copies, but differ
    when the model represents a different likelihood.

    Note: The key returned here considers only the spec and interpolation codes. All
    other model configuration options leave it unchanged (e.g. ``poi_name``, overriding
    parameter bounds, etc.).

    Args:
        model (pyhf.model.Model): model to generate a key for

    Returns:
        Tuple[str, Tuple[Tuple[str, str], ...]]: a key that identifies the model
        by its spec and interpcodes
    """
    interpcodes = []
    for mod_type in sorted(model.config.modifier_settings.keys()):
        code = model.config.modifier_settings[mod_type]["interpcode"]
        interpcodes.append((mod_type, code))
    # sort since different orderings result in equivalent models,
    # but distinct strings
    spec_str = json.dumps(model.spec, sort_keys=True)
    return (spec_str, tuple(interpcodes))


def yield_stdev(
    model: pyhf.pdf.Model,
    parameters: np.ndarray,
    uncertainty: np.ndarray,
    corr_mat: np.ndarray,
) -> Tuple[List[List[float]], List[float]]:
    """Calculates symmetrized yield standard deviation of a model, per bin and channel.

    Returns both the uncertainties per bin (in a list of channels), and the uncertainty
    of the total yield per channel (again, for a list of channels). To calculate the
    uncertainties for the total yield, the function internally treats the sum of yields
    per channel like another channel with one bin. The results of this function are
    cached to speed up subsequent calls with the same arguments.

    Args:
        model (pyhf.pdf.Model): the model for which to calculate the standard deviations
            for all bins
        parameters (np.ndarray): central values of model parameters
        uncertainty (np.ndarray): uncertainty of model parameters
        corr_mat (np.ndarray): correlation matrix

    Returns:
        Tuple[List[List[float]], List[float]]:
            - list of channels, each channel is a list of standard deviations per bin
            - list of standard deviations per channel
    """
    # check whether results are already stored in cache
    cached_results = _YIELD_STDEV_CACHE.get(
        (
            _hashable_model_key(model),
            tuple(parameters),
            tuple(uncertainty),
            corr_mat.data.tobytes(),
        ),
        None,
    )
    if cached_results is not None:
        # return results from cache
        return cached_results

    # the lists up_variations and down_variations will contain the model distributions
    # with all parameters varied individually within uncertainties
    # indices: variation, channel, bin
    # following the channels contained in the model, there are additional entries with
    # yields summed per channel (internally treated like additional channels) to get the
    # per-channel uncertainties
    up_variations = []
    down_variations = []

    # calculate the model distribution for every parameter varied up and down
    # within the respective uncertainties
    for i_par in range(model.config.npars):
        # central parameter values, but one parameter varied within uncertainties
        up_pars = parameters.copy().astype(float)  # ensure float for correct addition
        up_pars[i_par] += uncertainty[i_par]
        down_pars = parameters.copy().astype(float)
        down_pars[i_par] -= uncertainty[i_par]

        # total model distribution with this parameter varied up
        up_comb = pyhf.tensorlib.to_numpy(
            model.expected_data(up_pars, include_auxdata=False)
        )
        # turn into list of channels
        up_yields = [
            up_comb[model.config.channel_slices[ch]] for ch in model.config.channels
        ]
        # append list of yields summed per channel
        up_yields += [np.asarray([sum(chan_yields)]) for chan_yields in up_yields]
        up_variations.append(up_yields)

        # total model distribution with this parameter varied down
        down_comb = pyhf.tensorlib.to_numpy(
            model.expected_data(down_pars, include_auxdata=False)
        )
        # turn into list of channels
        down_yields = [
            down_comb[model.config.channel_slices[ch]] for ch in model.config.channels
        ]
        # append list of yields summed per channel
        down_yields += [np.asarray([sum(chan_yields)]) for chan_yields in down_yields]
        down_variations.append(down_yields)

    # convert to awkward arrays for further processing
    up_variations_ak = ak.from_iter(up_variations)
    down_variations_ak = ak.from_iter(down_variations)

    # calculate symmetric uncertainties for all components
    sym_uncs = (up_variations_ak - down_variations_ak) / 2

    # calculate total variance, indexed by channel and bin (per-channel numbers act like
    # additional channels with one bin each)
    if np.count_nonzero(corr_mat - np.diagflat(np.ones_like(parameters))) == 0:
        # no off-diagonal contributions from correlation matrix (e.g. pre-fit)
        total_variance = np.sum(np.power(sym_uncs, 2), axis=0)
    else:
        # full calculation including off-diagonal contributions
        # with v as vector of variations (each element contains yields under variation)
        # and M as correlation matrix, calculate variance as follows:
        # variance = sum_i sum_j v[i] * M[i, j] * v[j]
        # where the product between elements of v again is elementwise (multiplying bin
        # yields), and the final variance shape is the same as element of v (yield
        # uncertainties per bin and per channel)

        # possible optimizations that could be considered here:
        #   - skipping staterror-staterror terms for per-bin calculation (orthogonal)
        #   - taking advantage of correlation matrix symmetry
        #   - (optional) skipping combinations with correlations below threshold

        # calculate M[i, j] * v[j] first, indices: pars (i), pars (j), channel, bin
        m_times_v = corr_mat[..., np.newaxis, np.newaxis] * sym_uncs[np.newaxis, ...]
        # now multiply by v[i] as well, indices: pars(i), pars(j), channel, bin
        v_times_m_times_v = sym_uncs[:, np.newaxis, ...] * m_times_v
        # finally perform sums over i and j, remaining indices: channel, bin
        # ak.flatten due to https://github.com/scikit-hep/awkward-1.0/issues/1283
        assert v_times_m_times_v.ndim == 4, (
            "unexpected dimensionality of array, please report if you run into this: "
            "https://github.com/scikit-hep/cabinetry/issues/326"
        )
        total_variance = np.sum(ak.flatten(v_times_m_times_v, axis=1), axis=0)

    # convert to standard deviations per bin and per channel
    n_channels = len(model.config.channels)
    total_stdev_per_bin = np.sqrt(total_variance[:n_channels])
    total_stdev_per_channel = ak.flatten(np.sqrt(total_variance[n_channels:]))
    log.debug(f"total stdev is {total_stdev_per_bin}")
    log.debug(f"total stdev per channel is {total_stdev_per_channel}")

    # convert to lists
    total_stdev_per_bin = ak.to_list(total_stdev_per_bin)
    total_stdev_per_channel = ak.to_list(total_stdev_per_channel)

    # save to cache
    _YIELD_STDEV_CACHE.update(
        {
            (
                _hashable_model_key(model),
                tuple(parameters),
                tuple(uncertainty),
                corr_mat.data.tobytes(),
            ): (
                total_stdev_per_bin,
                total_stdev_per_channel,
            )
        }
    )

    return total_stdev_per_bin, total_stdev_per_channel


def prediction(
    model: pyhf.pdf.Model,
    *,
    fit_results: Optional[FitResults] = None,
    label: Optional[str] = None,
) -> ModelPrediction:
    """Returns model prediction, including model yields and uncertainties.

    If the optional fit result is not provided, the pre-fit Asimov yields and
    uncertainties are calculated. If the fit result is provided, the best-fit parameter
    values, uncertainties, and the parameter correlations are used to obtain the post-
    fit model and its associated uncertainties.

    Args:
        model (pyhf.pdf.Model): model to evaluate yield prediction for
        fit_results (Optional[FitResults], optional): parameter configuration to use,
            includes best-fit settings and uncertainties, as well as correlation matrix,
            defaults to None (then the pre-fit configuration is used)
        label (Optional[str], optional): label to include in model prediction, defaults
            to None (then will use "pre-fit" if fit results are not included, and "post-
            fit" otherwise)

    Returns:
        ModelPrediction: model, yields and uncertainties per bin and channel
    """
    if fit_results is not None:
        if fit_results.labels != model.config.par_names():
            log.warning("parameter names in fit results and model do not match")
        # fit results specified, so they are used
        param_values = fit_results.bestfit
        param_uncertainty = fit_results.uncertainty
        corr_mat = fit_results.corr_mat
        label = "post-fit" if label is None else label

    else:
        # no fit results specified, generate pre-fit parameter values, uncertainties,
        # and diagonal correlation matrix
        param_values = asimov_parameters(model)
        param_uncertainty = prefit_uncertainties(model)
        corr_mat = np.zeros(shape=(len(param_values), len(param_values)))
        np.fill_diagonal(corr_mat, 1.0)
        label = "pre-fit" if label is None else label

    yields_combined = pyhf.tensorlib.to_numpy(
        model.main_model.expected_data(param_values, return_by_sample=True)
    )  # all channels concatenated

    # slice the yields into list of lists (of lists) where first index is channel,
    # second index is sample (and third index is bin)
    model_yields = [
        yields_combined[:, model.config.channel_slices[ch]].tolist()
        for ch in model.config.channels
    ]

    # calculate the total standard deviation of the model prediction
    # indices: channel (and bin) for per-bin uncertainties, channel for per-channel
    total_stdev_model_bins, total_stdev_model_channels = yield_stdev(
        model, param_values, param_uncertainty, corr_mat
    )

    return ModelPrediction(
        model, model_yields, total_stdev_model_bins, total_stdev_model_channels, label
    )


def unconstrained_parameter_count(model: pyhf.pdf.Model) -> int:
    """Returns the number of unconstrained parameters in a model.

    The number is the sum of all independent parameters in a fit. A shapefactor that
    affects multiple bins enters the count once for each independent bin. Parameters
    that are set to constant are not included in the count.

    Args:
        model (pyhf.pdf.Model): model to count parameters for

    Returns:
        int: number of unconstrained parameters
    """
    n_pars = 0
    for parname in model.config.par_order:
        if (
            not model.config.param_set(parname).constrained
            and not model.config.param_set(parname).suggested_fixed
        ):
            n_pars += model.config.param_set(parname).n_parameters
    return n_pars


def _parameter_index(par_name: str, labels: Union[List[str], Tuple[str, ...]]) -> int:
    """Returns the position of a parameter with a given name in the list of parameters.

    Useful together with ``pyhf.pdf._ModelConfig.par_names`` to find the position of a
    parameter when the name is known. If the parameter is not found, logs an error and
    returns a default value of -1.

    Args:
        par_name (str): name of parameter to find in list
        labels (Union[List[str], Tuple[str, ...]]): list or tuple with all parameter
            names in the model

    Returns:
        int: index of parameter
    """
    par_index = next((i for i, label in enumerate(labels) if label == par_name), -1)
    if par_index == -1:
        log.error(f"parameter {par_name} not found in model")
    return par_index


def _strip_auxdata(model: pyhf.pdf.Model, data: List[float]) -> List[float]:
    """Always returns observed yields, no matter whether data includes auxdata.

    Args:
        model (pyhf.pdf.Model): model to which data corresponds to
        data (List[float]): data, either including auxdata which is then stripped off or
            only observed yields

    Returns:
        List[float]: observed data yields
    """
    n_bins_total = sum(model.config.channel_nbins.values())
    if len(data) != n_bins_total:
        # strip auxdata, only observed yields are needed
        data = data[:n_bins_total]
    return data


def _data_per_channel(model: pyhf.pdf.Model, data: List[float]) -> List[List[float]]:
    """Returns data split per channel, and strips off auxiliary data if included.

    Args:
        model (pyhf.pdf.Model): model to which data corresponds to
        data (List[float]): data (not split by channel), can either include auxdata
            which is then stripped off, or only observed yields

    Returns:
        List[List[float]]: data per channel and per bin
    """
    # strip off auxiliary data
    data_combined = _strip_auxdata(model, data)

    # data is indexed by channel (and bin)
    data_yields = [
        data_combined[model.config.channel_slices[ch]] for ch in model.config.channels
    ]
    return data_yields


def _filter_channels(
    model: pyhf.pdf.Model, channels: Optional[Union[str, List[str]]]
) -> List[str]:
    """Returns a list of channels in a model after applying filtering.

    Args:
        model (pyhf.pdf.Model): model from which to extract channels
        channels (Optional[Union[str, List[str]]]): name of channel or list of channels
            to filter, only including those channels provided via this argument in the
            return of the function

    Returns:
        List[str]: list of channels after filtering
    """
    # channels included in model
    filtered_channels = model.config.channels
    # if one or more custom channels are provided, only include those
    if channels is not None:
        if isinstance(channels, str):
            channels = [channels]  # ensure list
        # only include if channel exists in model
        filtered_channels = [ch for ch in channels if ch in model.config.channels]

    if filtered_channels == []:
        log.warning(
            f"channel(s) {channels} not found in model, available channel(s): "
            f"{model.config.channels}"
        )

    return filtered_channels


def match_fit_results(model: pyhf.pdf.Model, fit_results: FitResults) -> FitResults:
    """Matches results from a fit to a model by adding or removing parameters as needed.

    If the fit results contain parameters missing in the model, these parameters are not
    included in the returned fit results. If the fit results do not include parameters
    used in the model, they are added to the fit results. The best-fit value for such
    parameters are the Asimov values as returned by ``asimov_parameters`` (initial
    parameter settings for unconstrained parameters), and the associated uncertainties
    as given by ``prefit_uncertainties`` (zero uncertainty for unconstrained or fixed
    parameters). These parameters furthermore are assumed to have no correlation with
    any other parameters. If required, parameters are re-ordered to match the target
    model.

    Args:
        model (pyhf.pdf.Model): model to match fit results to
        fit_results (FitResults): fit results to be updated in order to match model

    Returns:
        FitResults: fit results matching the model
    """
    # start with the assumption that the provided fit results contain no relevant
    # information for the target model at all, and initialize everything accordingly
    # then inject information contained in provided fit results and return the new
    # fit results matching the target model at the end

    bestfit = asimov_parameters(model)  # Asimov parameter values for target model
    uncertainty = prefit_uncertainties(model)  # pre-fit uncertainties for target model
    labels = model.config.par_names()  # labels for target model

    # indices of parameters in current fit results, or None if they are missing
    indices_for_corr: List[Optional[int]] = [None] * len(labels)

    # loop over all required parameters
    for target_idx, target_label in enumerate(labels):
        # if parameters are missing in fit results, no further action is needed here
        # as all relevant objects have been initialized with that assumption
        if target_label in fit_results.labels:
            # fit results contain parameter, find its position
            idx_in_fit_results = fit_results.labels.index(target_label)
            # override pre-fit values by actual fit results for parameters
            bestfit[target_idx] = fit_results.bestfit[idx_in_fit_results]
            uncertainty[target_idx] = fit_results.uncertainty[idx_in_fit_results]
            # update indices for correlation matrix reconstruction
            indices_for_corr[target_idx] = idx_in_fit_results

    # re-build correlation matrix: start with diagonal matrix (assuming fit results do
    # not contain relevant info), and then insert values provided in fit results
    corr_mat = np.diagflat(np.ones_like(labels, dtype=float))
    for i_target, i_orig in enumerate(indices_for_corr):
        for j_target, j_orig in enumerate(indices_for_corr):
            # i_target and j_target are positions in matched correlation matrix
            # i_orig and j_orig are positions in old matrix, None if missing from there
            if i_orig is not None and j_orig is not None:
                # if i_orig or j_orig are None, one of the parameters are not part of
                # original fit result, and no update to correlation matrix is needed
                corr_mat[i_target][j_target] = fit_results.corr_mat[i_orig][j_orig]

    fit_results_matched = FitResults(
        np.asarray(bestfit),
        np.asarray(uncertainty),
        labels,
        corr_mat,
        fit_results.best_twice_nll,
        fit_results.goodness_of_fit,
    )
    return fit_results_matched



from collections import namedtuple
import copy
import pathlib
from unittest import mock

import matplotlib.figure
import numpy as np
import pyhf
import pytest

from cabinetry import fit
from cabinetry import model_utils
from cabinetry import visualize


MockHistogram = namedtuple("MockHistogram", ["bins", "yields", "stdev"])


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (("SR", "pre-fit"), "SR_prefit.pdf"),
        (("SR", "post-fit"), "SR_postfit.pdf"),
        (("SR 1", "prefit"), "SR-1_prefit.pdf"),
        (("SR 1", "postfit"), "SR-1_postfit.pdf"),
    ],
)
def test__figure_name(test_input, expected):
    assert visualize._figure_name(*test_input) == expected


def test__total_yield_uncertainty():
    stdev_list = [np.asarray([0.1, 0.2, 0.1]), np.asarray([0.3, 0.2, 0.1])]
    expected_uncertainties = [0.31622777, 0.28284271, 0.14142136]
    assert np.allclose(
        visualize._total_yield_uncertainty(stdev_list), expected_uncertainties
    )


@mock.patch("cabinetry.visualize._total_yield_uncertainty", return_value=[0.2])
@mock.patch(
    "cabinetry.visualize.plot_model.data_mc", return_value=matplotlib.figure.Figure()
)
@mock.patch(
    "cabinetry.histo.Histogram.from_config",
    return_value=MockHistogram([0.0, 1.0], [1.0], [0.1]),
)
def test_data_mc_from_histograms(mock_load, mock_draw, mock_stdev):
    config = {
        "General": {"HistogramFolder": "tmp_hist"},
        "Regions": [{"Name": "reg_1", "Variable": "x"}],
        "Samples": [{"Name": "sample_1"}, {"Name": "data", "Data": True}],
    }
    figure_folder = pathlib.Path("tmp")
    histogram_folder = pathlib.Path("tmp_hist")

    fig_dict_list = visualize.data_mc_from_histograms(
        config, figure_folder=figure_folder
    )
    assert len(fig_dict_list) == 1
    assert isinstance(fig_dict_list[0]["figure"], matplotlib.figure.Figure)
    assert fig_dict_list[0]["region"] == "reg_1"

    # the call_args_list contains calls (outer round brackets), first filled with
    # arguments (inner round brackets) and then keyword arguments
    assert mock_load.call_args_list == [
        (
            (
                histogram_folder,
                {"Name": "reg_1", "Variable": "x"},
                {"Name": "data", "Data": True},
                {},
            ),
            {"modified": True},
        ),
        (
            (
                histogram_folder,
                {"Name": "reg_1", "Variable": "x"},
                {"Name": "sample_1"},
                {},
            ),
            {"modified": True},
        ),
    ]
    assert mock_stdev.call_args_list == [(([[0.1]],), {})]
    assert mock_draw.call_args_list == [
        (
            (
                [
                    {"label": "data", "isData": True, "yields": [1.0], "variable": "x"},
                    {
                        "label": "sample_1",
                        "isData": False,
                        "yields": [1.0],
                        "variable": "x",
                    },
                ],
                [0.2],
                [0.0, 1.0],
            ),
            {
                "figure_path": figure_folder / "reg_1_prefit.pdf",
                "log_scale": None,
                "log_scale_x": False,
                "label": "reg_1\npre-fit",
                "close_figure": False,
            },
        )
    ]

    # custom log scale settings, close figure, do not save figure
    _ = visualize.data_mc_from_histograms(
        config,
        figure_folder=figure_folder,
        log_scale=True,
        log_scale_x=True,
        close_figure=True,
        save_figure=False,
    )
    assert mock_draw.call_args[1] == {
        "figure_path": None,
        "log_scale": True,
        "log_scale_x": True,
        "label": "reg_1\npre-fit",
        "close_figure": True,
    }


@mock.patch(
    "cabinetry.visualize.plot_model.data_mc", return_value=matplotlib.figure.Figure()
)
@mock.patch("cabinetry.templates.builder._binning", return_value=np.asarray([1, 2]))
@mock.patch(
    "cabinetry.configuration.region_dict",
    side_effect=[
        {"Name": "region", "Variable": "x", "Binning": [1, 2]},
        {"Name": "region"},
    ],
)
@mock.patch(
    "cabinetry.model_utils._filter_channels",
    side_effect=[["Signal Region"], ["Signal Region"], ["Signal Region"], []],
)
@mock.patch("cabinetry.model_utils._data_per_channel", return_value=[[12.0]])
def test_data_mc(mock_data, mock_filter, mock_dict, mock_bins, mock_draw, example_spec):
    config = {"config": "abc"}
    figure_folder = "tmp"
    model = pyhf.Workspace(example_spec).model()
    model_pred = model_utils.ModelPrediction(
        model, [[[10.0]]], [[0.3]], [0.3], "pre-fit"
    )
    data = [12.0, 1.0]

    # pre-fit plot
    fig_dict_list = visualize.data_mc(
        model_pred, data, config=config, figure_folder=figure_folder
    )
    assert len(fig_dict_list) == 1
    assert isinstance(fig_dict_list[0]["figure"], matplotlib.figure.Figure)
    assert fig_dict_list[0]["region"] == "Signal Region"

    assert mock_data.call_args_list == [((model, data), {})]
    assert mock_filter.call_args_list == [((model, None), {})]
    assert mock_dict.call_args_list == [((config, "Signal Region"), {})]
    assert mock_bins.call_args_list == [
        (({"Name": "region", "Variable": "x", "Binning": [1, 2]},), {})
    ]

    expected_histograms = [
        {
            "label": "Signal",
            "isData": False,
            "yields": np.asarray([10.0]),
            "variable": "x",
        },
        {
            "label": "Data",
            "isData": True,
            "yields": np.asarray([12.0]),
            "variable": "x",
        },
    ]
    assert mock_draw.call_count == 1
    assert mock_draw.call_args_list[0][0][0] == expected_histograms
    assert np.allclose(mock_draw.call_args_list[0][0][1], np.asarray([0.3]))
    np.testing.assert_equal(mock_draw.call_args_list[0][0][2], np.asarray([1, 2]))
    assert mock_draw.call_args_list[0][1] == {
        "figure_path": pathlib.Path("tmp/Signal-Region_prefit.pdf"),
        "log_scale": None,
        "log_scale_x": False,
        "label": "Signal Region\npre-fit",
        "close_figure": False,
    }

    # post-fit plot (different label in model prediction), custom scale, close figure,
    # do not save figure, histogram input mode: no binning or variable specified (via
    # side effect)
    model_pred = model_utils.ModelPrediction(
        model, [[[11.0]]], [[0.2]], [0.2], "post-fit"
    )
    _ = visualize.data_mc(
        model_pred,
        data,
        config=config,
        figure_folder=figure_folder,
        log_scale=False,
        close_figure=True,
        save_figure=False,
    )

    assert mock_draw.call_count == 2
    # yield at best-fit point is different from pre-fit
    assert np.allclose(mock_draw.call_args[0][0][0]["yields"], 11.0)
    assert np.allclose(mock_draw.call_args[0][1], np.asarray([0.2]))
    # observable defaults to "bin"
    assert mock_draw.call_args[0][0][0]["variable"] == "bin"
    assert mock_draw.call_args[0][0][1]["variable"] == "bin"
    # binning falls back to default
    np.testing.assert_equal(mock_draw.call_args[0][2], np.asarray([0, 1]))
    assert mock_draw.call_args[1] == {
        "figure_path": None,  # figure not saved
        "log_scale": False,
        "log_scale_x": False,
        "label": "Signal Region\npost-fit",
        "close_figure": True,
    }

    # no config specified, default variable name and bin edges
    _ = visualize.data_mc(model_pred, data)
    assert mock_draw.call_args[0][0][0]["variable"] == "bin"
    assert mock_draw.call_args[0][0][1]["variable"] == "bin"
    assert mock_draw.call_args[0][0][1]["yields"] == np.asarray(data[:1])
    np.testing.assert_equal(mock_draw.call_args[0][2], np.asarray([0, 1]))

    # no matching channels (via side_effect)
    assert visualize.data_mc(model_pred, data, channels="abc") is None
    assert mock_filter.call_args == ((model, "abc"), {})
    assert mock_draw.call_count == 3  # no new call


@mock.patch(
    "cabinetry.histo.Histogram.from_path",
    side_effect=[
        MockHistogram([0.0, 1.0], [2.0], [0.2]),
        MockHistogram([0.0, 1.0], [3.0], [0.3]),
        MockHistogram([0.0, 1.0], [4.0], [0.4]),
        MockHistogram([0.0, 1.0], [5.0], [0.5]),
    ]
    * 3,
)
@mock.patch(
    "cabinetry.histo.Histogram.from_config",
    return_value=MockHistogram([0.0, 1.0], [1.0], [0.1]),
)
@mock.patch(
    "cabinetry.visualize.plot_model.templates", return_value=matplotlib.figure.Figure()
)
def test_templates(mock_draw, mock_histo_config, mock_histo_path, tmp_path):
    # the side effects are repeated for the patched Histogram.from_path
    # to check all relevant behavior (including the unknown backend check)
    nominal_path = tmp_path / "region_sample_Nominal_modified.npz"
    up_path = tmp_path / "region_sample_sys_Up_modified.npz"
    down_path = tmp_path / "region_sample_sys_Down_modified.npz"
    region = {"Name": "region", "Variable": "x"}
    sample = {"Name": "sample"}
    config = {
        "General": {"HistogramFolder": tmp_path},
        "Regions": [region],
        "Samples": [sample, {"Name": "data", "Data": True}],
        "Systematics": [{"Name": "sys"}],
    }

    folder_path = "tmp"
    figure_path = pathlib.Path(folder_path) / "templates/region_sample_sys.pdf"

    # add fake histograms for glob
    nominal_path.touch()
    up_path.touch()
    down_path.touch()

    # also add a file that matches pattern but is not needed
    (tmp_path / "region_sample_sys_unknown_modified.npz").touch()

    fig_dict_list = visualize.templates(config, figure_folder=folder_path)
    assert len(fig_dict_list) == 1
    assert isinstance(fig_dict_list[0]["figure"], matplotlib.figure.Figure)
    assert fig_dict_list[0]["region"] == "region"
    assert fig_dict_list[0]["sample"] == "sample"
    assert fig_dict_list[0]["systematic"] == "sys"

    # nominal histogram loading
    assert mock_histo_config.call_args_list == [((tmp_path, region, sample, {}), {})]
    # variation histograms
    down_path_orig = pathlib.Path(str(down_path).replace("_modified", ""))
    up_path_orig = pathlib.Path(str(up_path).replace("_modified", ""))
    assert mock_histo_path.call_args_list == [
        ((down_path_orig,), {}),
        ((down_path,), {}),
        ((up_path_orig,), {}),
        ((up_path,), {}),
    ]

    nominal = {"yields": [1.0], "stdev": [0.1]}
    up_orig = {"yields": [4.0], "stdev": [0.4]}
    down_orig = {"yields": [2.0], "stdev": [0.2]}
    up_mod = {"yields": [5.0], "stdev": [0.5]}
    down_mod = {"yields": [3.0], "stdev": [0.3]}
    bins = [0.0, 1.0]
    assert mock_draw.call_args_list == [
        (
            (nominal, up_orig, down_orig, up_mod, down_mod, bins, "x"),
            {
                "figure_path": figure_path,
                "label": "region: region\nsample: sample\nsystematic: sys",
                "close_figure": False,
            },
        )
    ]

    # close figure, do not save figure, and remove variable information from config
    # (simulating histogram inputs), so variable defaults to "observable"
    histo_config = copy.deepcopy(config)
    histo_config["Regions"] = [{"Name": "region"}]
    _ = visualize.templates(
        histo_config, figure_folder=folder_path, close_figure=True, save_figure=False
    )
    assert mock_draw.call_args == (
        (nominal, up_orig, down_orig, up_mod, down_mod, bins, "observable"),
        {
            "figure_path": None,
            "label": "region: region\nsample: sample\nsystematic: sys",
            "close_figure": True,
        },
    )

    # remove files for variation histograms
    up_path.unlink()
    down_path.unlink()

    assert mock_draw.call_count == 2  # two calls so far
    _ = visualize.templates(config, figure_folder=folder_path)
    assert mock_draw.call_count == 2  # no new call, since no variations found

    # no systematics in config
    config = {
        "General": {"HistogramFolder": tmp_path},
        "Regions": [region],
        "Samples": [sample, {"Name": "data", "Data": True}],
    }
    _ = visualize.templates(config, figure_folder=folder_path)
    assert mock_draw.call_count == 2  # no systematics, so no new calls


@mock.patch(
    "cabinetry.visualize.plot_result.correlation_matrix",
    return_value=matplotlib.figure.Figure(),
)
def test_correlation_matrix(mock_draw):
    corr_mat = np.asarray([[1.0, 0.2, 0.1], [0.2, 1.0, 0.1], [0.1, 0.1, 1.0]])
    corr_mat_pruned = np.asarray([[1.0, 0.2], [0.2, 1.0]])
    labels = ["a", "b", "c"]
    labels_pruned = ["a", "b"]
    folder_path = "tmp"
    figure_path = pathlib.Path(folder_path) / "correlation_matrix.pdf"
    fit_results = fit.FitResults(np.empty(0), np.empty(0), labels, corr_mat, 1.0)

    # pruning with threshold
    fig = visualize.correlation_matrix(
        fit_results, figure_folder=folder_path, pruning_threshold=0.15
    )
    assert isinstance(fig, matplotlib.figure.Figure)

    mock_draw.assert_called_once()
    assert np.allclose(mock_draw.call_args[0][0], corr_mat_pruned)
    assert np.any(
        [mock_draw.call_args[0][1][i] == labels[i] for i in range(len(labels_pruned))]
    )
    assert mock_draw.call_args[1] == {"figure_path": figure_path, "close_figure": True}

    # pruning of fixed parameter (all zeros in correlation matrix row/column), do not
    # close figure, do not save
    corr_mat_fixed = np.asarray([[1.0, 0.2, 0.0], [0.2, 1.0, 0.0], [0.0, 0.0, 0.0]])
    fit_results_fixed = fit.FitResults(
        np.empty(0), np.empty(0), labels, corr_mat_fixed, 1.0
    )
    _ = visualize.correlation_matrix(
        fit_results_fixed,
        figure_folder=folder_path,
        close_figure=False,
        save_figure=False,
    )
    assert np.allclose(mock_draw.call_args_list[1][0][0], corr_mat_pruned)
    assert np.any(
        [
            mock_draw.call_args_list[1][0][1][i] == labels[i]
            for i in range(len(labels_pruned))
        ]
    )
    assert mock_draw.call_args[1] == {"figure_path": None, "close_figure": False}


@mock.patch(
    "cabinetry.visualize.plot_result.pulls", return_value=matplotlib.figure.Figure()
)
def test_pulls(mock_draw):
    bestfit = np.asarray([0.8, 1.0, 1.05, 1.1])
    uncertainty = np.asarray([0.9, 1.0, 0.03, 0.7])
    labels = ["a", "b", "staterror_region[0]", "c"]
    exclude = ["a"]
    folder_path = "tmp"
    fit_results = fit.FitResults(bestfit, uncertainty, labels, np.empty(0), 1.0)

    filtered_bestfit = np.asarray([1.0, 1.1])
    filtered_uncertainty = np.asarray([1.0, 0.7])
    filtered_labels = np.asarray(["b", "c"])
    figure_path = pathlib.Path(folder_path) / "pulls.pdf"

    # with filtering
    fig = visualize.pulls(fit_results, figure_folder=folder_path, exclude=exclude)
    assert isinstance(fig, matplotlib.figure.Figure)

    mock_draw.assert_called_once()
    assert np.allclose(mock_draw.call_args[0][0], filtered_bestfit)
    assert np.allclose(mock_draw.call_args[0][1], filtered_uncertainty)
    assert np.any(
        [
            mock_draw.call_args[0][2][i] == filtered_labels[i]
            for i in range(len(filtered_labels))
        ]
    )
    assert mock_draw.call_args[1] == {"figure_path": figure_path, "close_figure": True}

    # filtering single parameter instead of list
    _ = visualize.pulls(fit_results, figure_folder=folder_path, exclude=exclude[0])

    assert np.allclose(mock_draw.call_args[0][0], filtered_bestfit)
    assert np.allclose(mock_draw.call_args[0][1], filtered_uncertainty)
    assert np.any(
        [
            mock_draw.call_args[0][2][i] == filtered_labels[i]
            for i in range(len(filtered_labels))
        ]
    )

    # without filtering via list, but with staterror removal, fixed parameter removal,
    # not closing figure, not saving
    fit_results.uncertainty[0] = 0.0

    bestfit_expected = np.asarray([1.0, 1.1])
    uncertainty_expected = np.asarray([1.0, 0.7])
    labels_expected = ["b", "c"]
    visualize.pulls(
        fit_results, figure_folder=folder_path, close_figure=False, save_figure=False
    )

    assert np.allclose(mock_draw.call_args[0][0], bestfit_expected)
    assert np.allclose(mock_draw.call_args[0][1], uncertainty_expected)
    assert np.any(
        [
            mock_draw.call_args[0][2][i] == labels_expected[i]
            for i in range(len(labels_expected))
        ]
    )
    assert mock_draw.call_args[1] == {"figure_path": None, "close_figure": False}


@mock.patch(
    "cabinetry.visualize.plot_result.ranking", return_value=matplotlib.figure.Figure()
)
def test_ranking(mock_draw):
    bestfit = np.asarray([1.2, 0.1])
    uncertainty = np.asarray([0.2, 0.8])
    labels = ["staterror_a", "modeling"]
    impact_prefit_up = np.asarray([0.1, 0.5])
    impact_prefit_down = np.asarray([-0.2, -0.4])
    impact_postfit_up = np.asarray([0.1, 0.4])
    impact_postfit_down = np.asarray([-0.2, -0.3])
    ranking_results = fit.RankingResults(
        bestfit,
        uncertainty,
        labels,
        impact_prefit_up,
        impact_prefit_down,
        impact_postfit_up,
        impact_postfit_down,
    )
    folder_path = "tmp"

    figure_path = pathlib.Path(folder_path) / "ranking.pdf"
    bestfit_expected = np.asarray([0.1, 1.2])
    uncertainty_expected = np.asarray([0.8, 0.2])
    labels_expected = ["modeling", "staterror_a"]

    fig = visualize.ranking(ranking_results, figure_folder=folder_path)
    assert isinstance(fig, matplotlib.figure.Figure)

    assert mock_draw.call_count == 1
    assert np.allclose(mock_draw.call_args[0][0], bestfit_expected)
    assert np.allclose(mock_draw.call_args[0][1], uncertainty_expected)
    for i_lab, label in enumerate(mock_draw.call_args[0][2]):
        assert label == labels_expected[i_lab]
    assert np.allclose(mock_draw.call_args[0][3], impact_prefit_up[::-1])
    assert np.allclose(mock_draw.call_args[0][4], impact_prefit_down[::-1])
    assert np.allclose(mock_draw.call_args[0][5], impact_postfit_up[::-1])
    assert np.allclose(mock_draw.call_args[0][6], impact_postfit_down[::-1])
    assert mock_draw.call_args[1] == {"figure_path": figure_path, "close_figure": True}

    # maximum parameter amount specified, do not close figure, do not save figure
    _ = visualize.ranking(
        ranking_results,
        figure_folder=folder_path,
        max_pars=1,
        close_figure=False,
        save_figure=False,
    )
    assert mock_draw.call_count == 2
    assert np.allclose(mock_draw.call_args[0][0], bestfit_expected[0])
    assert np.allclose(mock_draw.call_args[0][1], uncertainty_expected[0])
    assert mock_draw.call_args[0][2] == labels_expected[0]
    assert np.allclose(mock_draw.call_args[0][3], impact_prefit_up[1])
    assert np.allclose(mock_draw.call_args[0][4], impact_prefit_down[1])
    assert np.allclose(mock_draw.call_args[0][5], impact_postfit_up[1])
    assert np.allclose(mock_draw.call_args[0][6], impact_postfit_down[1])
    assert mock_draw.call_args[1] == {"figure_path": None, "close_figure": False}


@mock.patch(
    "cabinetry.visualize.plot_result.scan", return_value=matplotlib.figure.Figure()
)
def test_scan(mock_draw):
    folder_path = "tmp"
    figure_path = pathlib.Path(folder_path) / "scan_a_0.pdf"

    par_name = "a[0]"
    par_mle = 1.5
    par_unc = 0.2
    par_vals = np.asarray([1.3, 1.5, 1.7])
    par_nlls = np.asarray([0.9, 0.0, 1.1])
    scan_results = fit.ScanResults(par_name, par_mle, par_unc, par_vals, par_nlls)

    fig = visualize.scan(scan_results, figure_folder=folder_path)
    assert isinstance(fig, matplotlib.figure.Figure)

    assert mock_draw.call_count == 1
    assert mock_draw.call_args[0][0] == par_name
    assert mock_draw.call_args[0][1] == par_mle
    assert mock_draw.call_args[0][2] == par_unc
    assert np.allclose(mock_draw.call_args[0][3], par_vals)
    assert np.allclose(mock_draw.call_args[0][4], par_nlls)
    assert mock_draw.call_args[1] == {"figure_path": figure_path, "close_figure": True}

    # do not close figure, do not save figure
    _ = visualize.scan(
        scan_results, figure_folder=folder_path, close_figure=False, save_figure=False
    )
    assert mock_draw.call_args[1] == {"figure_path": None, "close_figure": False}


@mock.patch(
    "cabinetry.visualize.plot_result.limit", return_value=matplotlib.figure.Figure()
)
def test_limit(mock_draw):
    folder_path = "tmp"
    figure_path = pathlib.Path(folder_path) / "limit.pdf"

    observed_CLs = np.asarray([0.75, 0.32, 0.02])
    expected_CLs = np.asarray([[0.1, 0.2, 0.3, 0.4, 0.5] for _ in range(3)])
    poi_values = np.asarray([0, 1, 2])
    confidence_level = 0.90
    limit_results = fit.LimitResults(
        3.0, np.empty(5), observed_CLs, expected_CLs, poi_values, confidence_level
    )

    fig = visualize.limit(limit_results, figure_folder=folder_path)
    assert isinstance(fig, matplotlib.figure.Figure)

    assert mock_draw.call_count == 1
    assert np.allclose(mock_draw.call_args[0][0], limit_results.observed_CLs)
    assert np.allclose(mock_draw.call_args[0][1], limit_results.expected_CLs)
    assert np.allclose(mock_draw.call_args[0][2], limit_results.poi_values)
    assert np.allclose(mock_draw.call_args[0][3], 1 - limit_results.confidence_level)
    assert mock_draw.call_args[1] == {"figure_path": figure_path, "close_figure": True}

    # do not close figure, do not save figure
    _ = visualize.limit(
        limit_results, figure_folder=folder_path, close_figure=False, save_figure=False
    )
    assert mock_draw.call_args[1] == {"figure_path": None, "close_figure": False}



"""The cabinetry library."""

import logging

import cabinetry.configuration  # noqa: F401
import cabinetry.fit  # noqa: F401
import cabinetry.model_utils  # noqa: F401
import cabinetry.route  # noqa: F401
import cabinetry.smooth  # noqa: F401
import cabinetry.tabulate  # noqa: F401
import cabinetry.templates  # noqa: F401
import cabinetry.visualize  # noqa: F401
import cabinetry.workspace  # noqa: F401


__version__ = "0.4.1"


def set_logging() -> None:
    """Sets up customized and verbose logging output.

    Logging can be alternatively customized with the Python ``logging`` module directly.
    """
    logging.basicConfig(format="%(levelname)s - %(name)s - %(message)s")
    logging.getLogger("cabinetry").setLevel(logging.DEBUG)
    logging.getLogger("pyhf").setLevel(logging.INFO)



"""Utility to import the Literal type."""

import sys

if sys.version_info < (3, 8):
    from typing_extensions import Literal  # noqa: F401, # pragma: no cover
else:
    from typing import Literal  # noqa: F401, # pragma: no cover
